import sys
import struct

class p01_segments:
	# Format: Checksum location, start address, end address, ... start address, end address
	os          = [0x000500, 0x000000, 0x0004FF, 0x000502, 0x003FFF, 0x020000, 0x07FFFD]
	engine_cal  = [0x008000, 0x008002, 0x013FFF]
	engine_diag = [0x014000, 0x014002, 0x016DFF]
	trans_cal   = [0x016E00, 0x016E02, 0x01BDFF]
	trans_diag  = [0x01BE00, 0x01BE02, 0x01C7FF]
	fuel_system = [0x01C800, 0x01C802, 0x01E51F]
	system      = [0x01E520, 0x01E522, 0x01EE9F]
	speedometer = [0x01EEA0, 0x01EEA2, 0x01EF9F]
	segments = [os, engine_cal, engine_diag, trans_cal, trans_diag, fuel_system, system, speedometer]
	segment_names = ["OS", "Engine Cal", "Engine Diag", "Trans Cal", "Trans Diag", "Fuel System", "System", "Speedometer"]

class p59_segments:
	# Format: Checksum location, start address, end address, ... start address, end address
	os          = [0x000500, 0x000000, 0x0004FF, 0x000502, 0x003FFF, 0x020000, 0x0FFFFD]
	engine_cal  = [0x008000, 0x008002, 0x015DFF]
	engine_diag = [0x015E00, 0x015E02, 0x018F8F]
	trans_cal   = [0x018F90, 0x018F92, 0x01D26F]
	trans_diag  = [0x01D270, 0x01D272, 0x01DBCF]
	fuel_system = [0x01DBD0, 0x01DBD2, 0x01F18F]
	system      = [0x01F190, 0x01F192, 0x01FB4F]
	speedometer = [0x01FB50, 0x01FB52, 0x01FD7F]
	segments = [os, engine_cal, engine_diag, trans_cal, trans_diag, fuel_system, system, speedometer]
	segment_names = ["Operating System", "Engine Calibration", "Engine Diagnostics", "Transmission Calibration", "Transmission Diagnostics", "Fuel System", "System", "Speedometer"]

class gen3_checksum:
	def __init__(self, binPath = ""):
		# Global variables
		self.p59 = False
		self.segmentDefinitionClass = p01_segments
		self.binPath = binPath
		self.osid = 0
		self.outPath = "./undefined.bin"
		self.outRequest = False
		
		# Initialize other globals not yet used here
		self.binaryChecksums = [0] * len(p01_segments.segments)
		self.calculatedChecksums = [0] * len(p01_segments.segments)
		
		
		# Load the binary into memory, I'll do this the lazy/inefficient way because who cares about 1M?
		file = open(self.binPath, mode="rb")
		self.binary = file.read()
		# Create a copy for the output buffer, if checksums need to be updated
		self.binaryCorrected = bytearray(self.binary)
		self.binLength = len(self.binary)
		file.close()
		
		
		# Check if file size is equal to 1M or ~512KB and set variables accordingly
		if (self.binLength == 1048576):
			self.p59 = True
			# Reset segment values from binary
			p59_segments.engine_cal 	= [self.read_uint32(0x514),self.read_uint32(0x514)+2,self.read_uint32(0x518)]
			p59_segments.engine_diag 	= [self.read_uint32(0x51c),self.read_uint32(0x51c)+2,self.read_uint32(0x520)]
			p59_segments.trans_cal 		= [self.read_uint32(0x524),self.read_uint32(0x524)+2,self.read_uint32(0x528)]
			p59_segments.trans_diag 	= [self.read_uint32(0x52c),self.read_uint32(0x52c)+2,self.read_uint32(0x530)]
			p59_segments.fuel_system 	= [self.read_uint32(0x534),self.read_uint32(0x534)+2,self.read_uint32(0x538)]
			p59_segments.system		 	= [self.read_uint32(0x53c),self.read_uint32(0x53c)+2,self.read_uint32(0x540)]
			p59_segments.speedometer 	= [self.read_uint32(0x544),self.read_uint32(0x544)+2,self.read_uint32(0x548)]
			self.segmentDefinitionClass = p59_segments
		elif (self.binLength == 524288):
			self.p01 = False
		else:
			print ("Error: File size is not expected. Invalid binary??")
			exit()
		self.outRequest = False
		
		# Read relevant information from binary
		self.osid = self.read_uint32(p59_segments.os[0]+4)
		
		# Calculate the checksums
		self.__populate_checksum_lists()
					
	def request_output_file(self, outPath):
		self.outPath = outPath
		self.outRequest = True
		
	def write_output(self):
		if (self.outRequest):
			print("Writing updated binary to:",self.outPath)
			newBin = open(self.outPath, "wb")
			# write to file
			newBin.write(self.binaryCorrected)
			newBin.close()
			
	def print_information(self):
		print("Loaded binary file:", self.binPath)
		if (self.p59):
			print("ECU type: P59")
		else:
			print ("ECU type: P01")
		print ("Binary size:", self.binLength,"bytes")
		print ("OS ID:", self.osid)
		print("")
		
	def analyze_all_segments(self):
		print ("Analyzing segments")
		failedChecksum = False
		for i in range(8):
			existing = self.binaryChecksums[i]
			calculated = self.calculatedChecksums[i]
			status = "[ PASS ]"
			if (existing != calculated):
				status = "[ FAIL ]"
				failedChecksum = True
			print (str.rjust(p01_segments.segment_names[i],len(max(p01_segments.segment_names))+3),":",status,"Bin Checksum:", hex(existing),"| Calculated Checksum:", hex(calculated))
		
		if (failedChecksum):
			print ("\n\rOne or more segments have an invalid checksum!")
		else:
			print ("\n\rAll segments have valid checksums.")
		
	
	
	def read_segment_checksum(self, num = 0):
		if (num > 7 or num < 0):
			print ("ERROR: Invalid segment number passed. Valid numbers are 0-7")
			return
		segmentClass = p01_segments
		
		if (self.p59):
			segmentClass = p59_segments
		
		if (num == 0):
			return self.read_uint16(segmentClass.os[0])
		elif (num == 1):
			return self.read_uint16(segmentClass.engine_cal[0])
		elif (num == 2):
			return self.read_uint16(segmentClass.engine_diag[0])
		elif (num == 3):
			return self.read_uint16(segmentClass.trans_cal[0])
		elif (num == 4):
			return self.read_uint16(segmentClass.trans_diag[0])
		elif (num == 5):
			return self.read_uint16(segmentClass.fuel_system[0])
		elif (num == 6):
			return self.read_uint16(segmentClass.system[0])
		elif (num == 7):
			return self.read_uint16(segmentClass.speedometer[0])
	
	def generate_segment_checksum(self, num = 0):
		if (num > 7 or num < 0):
			print ("ERROR: Invalid segment number passed. Valid numbers are 0-7")
			return
		
		segmentClass = p01_segments
		if (self.p59):
			segmentClass = p59_segments
			
		if (num == 0):
			return self.__generate_checksum_at_addr(segmentClass.os)
		elif (num == 1):
			return self.__generate_checksum_at_addr(segmentClass.engine_cal)
		elif (num == 2):
			return self.__generate_checksum_at_addr(segmentClass.engine_diag)
		elif (num == 3):
			return self.__generate_checksum_at_addr(segmentClass.trans_cal)
		elif (num == 4):
			return self.__generate_checksum_at_addr(segmentClass.trans_diag)
		elif (num == 5):
			return self.__generate_checksum_at_addr(segmentClass.fuel_system)
		elif (num == 6):
			return self.__generate_checksum_at_addr(segmentClass.system)
		elif (num == 7):
			return self.__generate_checksum_at_addr(segmentClass.speedometer)
		
	
	def __generate_checksum_at_addr(self, addressArray):
		arrayLen = len(addressArray)
		if ( arrayLen < 3):
			print ("Failed to analyze. Passed array doesn't contain enough addresses")
			return
		if ((arrayLen % 2) != 1):
			print ("Failed to analyze section, passed array doesn't contain start/end pairs")
			return
		
		arrayIndex = 1
		sum = 0
		while ((arrayIndex + 1) < arrayLen):
			startAddr = addressArray[arrayIndex]
			endAddr = addressArray[arrayIndex + 1]
			size = endAddr - startAddr + 1;
			
			sum = sum + self.__sum_ushorts(startAddr, endAddr)
			arrayIndex = arrayIndex + 2
		
		sum = sum & 0xFFFF
		return (65536 - sum) & 0xFFFF
	
	def read_segments(self):
		print("Reading the segment addresses from the binary. Starts at 0x514")
		print(" Engine Cal:", hex(self.read_uint32(0x514)), hex(self.read_uint32(0x518)))
		print("Engine Diag:", hex(self.read_uint32(0x51C)), hex(self.read_uint32(0x520)))
		print("  Trans Cal:", hex(self.read_uint32(0x524)), hex(self.read_uint32(0x528)))
		print(" Trans Diag:", hex(self.read_uint32(0x52C)), hex(self.read_uint32(0x530)))
		print("Fuel System:", hex(self.read_uint32(0x534)), hex(self.read_uint32(0x538)))
		print("     System:", hex(self.read_uint32(0x53C)), hex(self.read_uint32(0x540)))
		print("Speedometer:", hex(self.read_uint32(0x544)), hex(self.read_uint32(0x548)))
	
	def __sum_ushorts(self, startAddress, endAddress):
		startAddr = startAddress
		endAddr = endAddress + 1
		bytes = self.binary[startAddr:endAddr]
		output = struct.unpack(">"+("H" * (len(bytes) // 2)), bytes)
		sum = 0
		for i in output:
			sum = sum + i
		sum = sum & 0xFFFF
		return sum
		
	def read_uint16(self, addr):
		return struct.unpack(">H",self.binary[addr:(addr+2)])[0]
		
	def read_uint32(self, addr):
		return struct.unpack(">I",self.binary[addr:(addr+4)])[0]
	
	def __test_checksum(self, checksum, checksum2):
		return (checksum + checksum2) & 0xFFFF
		
	def __populate_checksum_lists(self):
		for i in range(len(p01_segments.segments)):
			self.binaryChecksums[i] = self.read_segment_checksum(i)
			self.calculatedChecksums[i] = self.generate_segment_checksum(i)
			if (self.binaryChecksums[i] != self.calculatedChecksums[i]):
				# If checksums don't match, let's update the checksum in the output file buffer
				checksumAddr = self.segmentDefinitionClass.segments[i][0]
				# I'm sure there's a better way to split these up, but I'm more familiar with C
				newChecksum = [0] * 2
				newChecksum[0] = (self.calculatedChecksums[i] & 0xFF)
				newChecksum[1] = (self.calculatedChecksums[i] & 0xFF00) >> 8
				# Push new values into the output buffer
				self.binaryCorrected[checksumAddr] = newChecksum[1]
				self.binaryCorrected[checksumAddr+1] = newChecksum[0]


#### Main function if ran directly from command line ###
if __name__ == '__main__':
	print ("GM Gen 3 P01 & P59 Checksum Checker/Updater")
	print ("Version: 0.1a, Released MMMMM DD, YYYY\n")
	if sys.version_info[0] < 3:
		raise Exception("Python 3 or a more recent version is required.")
		
	if (len(sys.argv) < 2):
		print ("Error: No binary file was passed as an argument.")
		print ("EX: python3 checksum.py ./input.bin; Checks input checksums")
		print ("EX: python3 checksum.py ./input.bin ./output.bin; Checks input checksums, and writes corrected checksums to a copy (output.bin)")
		exit()
    
	inputPath = sys.argv[1]
	binary = gen3_checksum(inputPath)
	if (len(sys.argv) > 2):
		print ("Opening with request to output file, for corrected checksums")
		binary.request_output_file(sys.argv[2])
	
	binary.print_information()
	binary.analyze_all_segments()
	binary.write_output()

